﻿#region License Information
/* HeuristicLab
 * Copyright (C) 2002-2016 Heuristic and Evolutionary Algorithms Laboratory (HEAL)
 *
 * This file is part of HeuristicLab.
 *
 * HeuristicLab is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * HeuristicLab is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with HeuristicLab. If not, see <http://www.gnu.org/licenses/>.
 */
#endregion

using HeuristicLab.Analysis;
using HeuristicLab.Collections;
using HeuristicLab.Common;
using HeuristicLab.Common.Resources;
using HeuristicLab.Core;
using HeuristicLab.Data;
using HeuristicLab.MainForm;
using HeuristicLab.Optimization;
using HeuristicLab.Persistence.Default.Xml;
using System;
using System.Collections.Generic;
using System.Drawing;
using System.IO;
using System.Linq;
using System.Threading;
using System.Threading.Tasks;
using RunCreationClient = HeuristicLab.Clients.OKB.RunCreation.RunCreationClient;
using SingleObjectiveOKBProblem = HeuristicLab.Clients.OKB.RunCreation.SingleObjectiveOKBProblem;
using SingleObjectiveOKBSolution = HeuristicLab.Clients.OKB.RunCreation.SingleObjectiveOKBSolution;

namespace HeuristicLab.OptimizationExpertSystem.Common {
  [Item("Knowledge Center", "Currently in experimental phase, an expert system that makes algorithm suggestions based on fitness landscape analysis features and an optimization knowledge base.")]
  [Creatable(CreatableAttribute.Categories.TestingAndAnalysis, Priority = 119)]
  public sealed class KnowledgeCenter : IContent {

    public string Filename { get; set; }

    public static new Image StaticItemImage {
      get { return VSImageLibrary.Library; }
    }

    private IntValue maximumEvaluations;
    public IntValue MaximumEvaluations {
      get { return maximumEvaluations; }
    }

    private RunCollection instanceRuns;
    public RunCollection InstanceRuns {
      get { return instanceRuns; }
    }

    private RunCollection seededRuns;
    public RunCollection SeededRuns {
      get { return seededRuns; }
    }

    private RunCollection knowledgeBase;
    public RunCollection KnowledgeBase {
      get { return knowledgeBase; }
    }

    private SingleObjectiveOKBProblem problem;
    public SingleObjectiveOKBProblem Problem {
      get { return problem; }
    }

    private ItemList<IAlgorithm> suggestedInstances;
    private ReadOnlyItemList<IAlgorithm> readOnlySuggestedInstances;
    public ReadOnlyItemList<IAlgorithm> SuggestedInstances {
      get { return readOnlySuggestedInstances; }
    }

    private RunCollection problemInstances;
    public RunCollection ProblemInstances {
      get { return problemInstances; }
    }

    private CheckedItemList<IScope> solutionSeedingPool;
    public CheckedItemList<IScope> SolutionSeedingPool {
      get { return solutionSeedingPool; }
    }

    private EnumValue<SeedingStrategyTypes> seedingStrategy;
    public EnumValue<SeedingStrategyTypes> SeedingStrategy {
      get { return seedingStrategy; }
    }
    
    private BidirectionalLookup<long, IRun> algorithmId2RunMapping;
    private BidirectionalDictionary<long, IAlgorithm> algorithmId2AlgorithmInstanceMapping;

    private Run currentInstance;

    private bool Maximization {
      get { return Problem != null && Problem.ProblemId >= 0 && ((IValueParameter<BoolValue>)Problem.MaximizationParameter).Value.Value; }
    }

    public KnowledgeCenter() {
      maximumEvaluations = new IntValue(10000);
      instanceRuns = new RunCollection();
      seededRuns = new RunCollection();
      knowledgeBase = new RunCollection();
      suggestedInstances = new ItemList<IAlgorithm>();
      readOnlySuggestedInstances = suggestedInstances.AsReadOnly();
      problemInstances = new RunCollection();
      problem = new SingleObjectiveOKBProblem();
      algorithmId2RunMapping = new BidirectionalLookup<long, IRun>();
      algorithmId2AlgorithmInstanceMapping = new BidirectionalDictionary<long, IAlgorithm>();
      solutionSeedingPool = new CheckedItemList<IScope>();
      seedingStrategy = new EnumValue<SeedingStrategyTypes>(SeedingStrategyTypes.NoSeeding);
      RegisterEventHandlers();
    }

    private void ProblemOnProblemChanged(object sender, EventArgs eventArgs) {
      // TODO: Potentially, knowledge base has to be re-downloaded
    }

    private void RegisterEventHandlers() {
      maximumEvaluations.ValueChanged += MaximumEvaluationsOnValueChanged;
      problem.ProblemChanged += ProblemOnProblemChanged;
      problem.Solutions.ItemsAdded += ProblemSolutionsChanged;
      problem.Solutions.ItemsReplaced += ProblemSolutionsChanged;
      problem.Solutions.ItemsRemoved += ProblemSolutionsChanged;
      problem.Solutions.CollectionReset += ProblemSolutionsChanged;
      instanceRuns.CollectionReset += InformationChanged;
      instanceRuns.ItemsAdded += InformationChanged;
      instanceRuns.ItemsRemoved += InformationChanged;
      instanceRuns.Reset += InformationChanged;
      instanceRuns.UpdateOfRunsInProgressChanged += InformationChanged;
      knowledgeBase.CollectionReset += InformationChanged;
      knowledgeBase.ItemsAdded += InformationChanged;
      knowledgeBase.ItemsRemoved += InformationChanged;
    }

    private void MaximumEvaluationsOnValueChanged(object sender, EventArgs eventArgs) {
      UpdateSuggestions();
    }

    private void ProblemSolutionsChanged(object sender, EventArgs e) {
      foreach (var sol in Problem.Solutions.Select(x => x.Solution).OfType<IScope>()) {
        if (!SolutionSeedingPool.Contains(sol))
          SolutionSeedingPool.Add(sol, false);
      }
    }

    private void InformationChanged(object sender, EventArgs e) {
      var runCollection = sender as RunCollection;
      if (runCollection != null && runCollection.UpdateOfRunsInProgress) return;
      UpdateSuggestions();
    }

    private void UpdateInstanceProjection() {
      currentInstance = null;
      var filter = new HashSet<string>();

      if (Problem.ProblemId != -1) {
        currentInstance = new Run { Name = Problem.Name };
        Problem.Problem.CollectParameterValues(currentInstance.Parameters);
        foreach (var c in RunCreationClient.Instance.GetCharacteristicValues(Problem.ProblemId)) {
          var key = "Characteristic." + c.Name;
          currentInstance.Results.Add(key, RunCreationClient.Instance.ConvertToItem(c));
          filter.Add(key);
        }
      }
      // TODO: There is a problem with missing values that has to be solved
      // The common set of characteristics among all problems may be too small to be useful
      // It has to be decided somehow which problem instances to include in order to obtain the set of features that is expressive and available
      var allCharacteristics = ProblemInstances.Select(x => new HashSet<string>(x.Results.Where(y => y.Key.StartsWith("Characteristic.", StringComparison.Ordinal)).Select(y => y.Key))).ToList();
      if (allCharacteristics.Count == 0) return;

      var commonCharacteristics = filter.Count > 0 ? filter : allCharacteristics[0];
      for (var i = 0; i < allCharacteristics.Count; i++)
        commonCharacteristics.IntersectWith(allCharacteristics[i]);

      if (commonCharacteristics.Count == 0) return;

      var instances = new Dictionary<string, double[]>();
      foreach (var i in ProblemInstances)
        instances[i.Name] = i.Results.Where(x => commonCharacteristics.Contains(x.Key)).OrderBy(x => x.Key).Select(x => (double)((dynamic)x.Value).Value).ToArray();
      if (currentInstance != null)
        instances[currentInstance.Name] = currentInstance.Results.Where(x => commonCharacteristics.Contains(x.Key)).OrderBy(x => x.Key).Select(x => (double)((dynamic)x.Value).Value).ToArray();

      var allValues = instances.SelectMany(x => x.Value).ToList();
      var avg = allValues.Average();
      var stdev = allValues.StandardDeviation();

      // normalize characteristic values by transforming them to their z-score
      foreach (var key in instances.Keys.ToList()) {
        var arr = instances[key];
        for (var i = 0; i < arr.Length; i++) {
          arr[i] = (arr[i] - avg) / stdev;
        }
      }

      var key2Idx = new BidirectionalDictionary<string, int>();
      foreach (var kvp in instances.Select((k, i) => new { Index = i, Key = k.Key }))
        key2Idx.Add(kvp.Key, kvp.Index);

      #region MDS
      Func<double[], double[], double> euclid = (a, b) => Math.Sqrt(a.Zip(b, (x, y) => (x - y)).Sum(x => x * x));
      var num = instances.Count;
      var matrix = new DoubleMatrix(num, num);
      for (var i = 0; i < num - 1; i++) {
        for (var j = i + 1; j < num; j++) {
          matrix[i, j] = matrix[j, i] = euclid(instances[key2Idx.GetBySecond(i)], instances[key2Idx.GetBySecond(j)]);
        }
      }

      var coords = MultidimensionalScaling.KruskalShepard(matrix);
      #endregion
      #region PCA
      var ds = new double[instances.Count, commonCharacteristics.Count];
      foreach (var instance in instances) {
        var arr = instance.Value;
        for (var feature = 0; feature < arr.Length; feature++)
          ds[key2Idx.GetByFirst(instance.Key), feature] = arr[feature];
      }

      int info;
      double[] s2;
      double[,] v;
      alglib.pcabuildbasis(ds, ds.GetLength(0), ds.GetLength(1), out info, out s2, out v);
      #endregion

      ProblemInstances.UpdateOfRunsInProgress = true;
      try {
        foreach (var instance in ProblemInstances) {
          double x = 0, y = 0;
          for (var feature = 0; feature < ds.GetLength(1); feature++) {
            x += ds[key2Idx.GetByFirst(instance.Name), feature] * v[feature, 0];
            y += ds[key2Idx.GetByFirst(instance.Name), feature] * v[feature, 1];
          }

          IItem item;
          if (instance.Results.TryGetValue("Projection.PCA.X", out item)) {
            ((DoubleValue)item).Value = x;
          } else instance.Results.Add("Projection.PCA.X", new DoubleValue(x));
          if (instance.Results.TryGetValue("Projection.PCA.Y", out item)) {
            ((DoubleValue)item).Value = y;
          } else instance.Results.Add("Projection.PCA.Y", new DoubleValue(y));

          if (instance.Results.TryGetValue("Projection.MDS.X", out item)) {
            ((DoubleValue)item).Value = coords[key2Idx.GetByFirst(instance.Name), 0];
          } else instance.Results.Add("Projection.MDS.X", new DoubleValue(coords[key2Idx.GetByFirst(instance.Name), 0]));
          if (instance.Results.TryGetValue("Projection.MDS.Y", out item)) {
            ((DoubleValue)item).Value = coords[key2Idx.GetByFirst(instance.Name), 1];
          } else instance.Results.Add("Projection.MDS.Y", new DoubleValue(coords[key2Idx.GetByFirst(instance.Name), 1]));
        }
      } finally { ProblemInstances.UpdateOfRunsInProgress = false; }

      if (currentInstance != null) {
        double x = 0, y = 0;
        for (var feature = 0; feature < ds.GetLength(1); feature++) {
          x += ds[key2Idx.GetByFirst(currentInstance.Name), feature] * v[feature, 0];
          y += ds[key2Idx.GetByFirst(currentInstance.Name), feature] * v[feature, 1];
        }

        IItem item;
        if (currentInstance.Results.TryGetValue("Projection.PCA.X", out item)) {
          ((DoubleValue)item).Value = x;
        } else currentInstance.Results.Add("Projection.PCA.X", new DoubleValue(x));
        if (currentInstance.Results.TryGetValue("Projection.PCA.Y", out item)) {
          ((DoubleValue)item).Value = y;
        } else currentInstance.Results.Add("Projection.PCA.Y", new DoubleValue(y));

        if (currentInstance.Results.TryGetValue("Projection.MDS.X", out item)) {
          ((DoubleValue)item).Value = coords[key2Idx.GetByFirst(currentInstance.Name), 0];
        } else currentInstance.Results.Add("Projection.MDS.X", new DoubleValue(coords[key2Idx.GetByFirst(currentInstance.Name), 0]));
        if (currentInstance.Results.TryGetValue("Projection.MDS.Y", out item)) {
          ((DoubleValue)item).Value = coords[key2Idx.GetByFirst(currentInstance.Name), 1];
        } else currentInstance.Results.Add("Projection.MDS.Y", new DoubleValue(coords[key2Idx.GetByFirst(currentInstance.Name), 1]));
      }
    }

    public Tuple<double, double> ProjectCurrentInstance(string projection) {
      if (currentInstance == null) return null;
      var xKey = "Projection." + projection + ".X";
      var yKey = "Projection." + projection + ".Y";
      if (!currentInstance.Results.ContainsKey(xKey) || !currentInstance.Results.ContainsKey(yKey))
        return null;
      var x = ((DoubleValue)currentInstance.Results[xKey]).Value;
      var y = ((DoubleValue)currentInstance.Results[yKey]).Value;
      return Tuple.Create(x, y);
    }

    private static readonly HashSet<string> InterestingValueNames = new HashSet<string>() {
      "QualityPerEvaluations", "Problem Name", "Problem Type", "Algorithm Name", "Algorithm Type", "Maximization", "BestKnownQuality"
    };

    public Task<ResultCollection> StartAlgorithmAsync(int index) {
      return StartAlgorithmAsync(index, CancellationToken.None);
    }

    public Task<ResultCollection> StartAlgorithmAsync(int index, CancellationToken cancellation) {
      var selectedInstance = suggestedInstances[index];
      var algorithmClone = (IAlgorithm)selectedInstance.Clone();
      var problemClone = Problem.CloneProblem() as ISingleObjectiveHeuristicOptimizationProblem;
      if (problemClone == null) throw new InvalidOperationException("Problem is not of type " + typeof(ISingleObjectiveHeuristicOptimizationProblem).FullName);
      // TODO: It is assumed the problem instance by default is configured using no preexisting solution creator
      var seedingStrategyLocal = SeedingStrategy.Value;
      if (seedingStrategyLocal != SeedingStrategyTypes.NoSeeding) {
        if (!SolutionSeedingPool.CheckedItems.Any()) throw new InvalidOperationException("There are no solutions selected for seeding.");
        // TODO: It would be necessary to specify the solution creator somewhere (property and GUI)
        var seedingCreator = problemClone.Operators.OfType<IPreexistingSolutionCreator>().FirstOrDefault();
        if (seedingCreator == null) throw new InvalidOperationException("The problem does not contain a solution creator that allows seeding.");
        seedingCreator.PreexistingSolutionsParameter.Value.Replace(SolutionSeedingPool.CheckedItems.Select(x => x.Value));
        seedingCreator.SampleFromPreexistingParameter.Value.Value = seedingStrategyLocal == SeedingStrategyTypes.SeedBySampling;
        // TODO: WHY!? WHY??!?
        ((dynamic)problemClone.SolutionCreatorParameter).Value = (dynamic)seedingCreator;
      }
      algorithmClone.Problem = problemClone;
      algorithmClone.Prepare(true);
      IParameter stopParam;
      var monitorStop = true;
      if (algorithmClone.Parameters.TryGetValue("MaximumEvaluations", out stopParam)) {
        var maxEvalParam = stopParam as IValueParameter<Data.IntValue>;
        if (maxEvalParam != null) {
          maxEvalParam.Value.Value = MaximumEvaluations.Value;
          monitorStop = false;
        }
      }

      // TODO: The following can be simplified when we have async implementation patterns for our algorithms:
      // TODO: The closures can be removed and replaced with private member methods
      var waitHandle = new AutoResetEvent(false);

      #region EventHandler closures
      EventHandler exeStateChanged = (sender, e) => {
        if (algorithmClone.ExecutionState == ExecutionState.Stopped) {
          lock (Problem.Solutions) {
            foreach (var solution in algorithmClone.Results.Where(x => x.Name.ToLower().Contains("solution")).Select(x => x.Value).OfType<IScope>()) {
              Problem.Solutions.Add(new SingleObjectiveOKBSolution(Problem.ProblemId) {
                Quality = solution.Variables.ContainsKey(Problem.Problem.Evaluator.QualityParameter.ActualName) ? ((DoubleValue)solution.Variables[Problem.Problem.Evaluator.QualityParameter.ActualName].Value).Value : double.NaN,
                Solution = (IItem)solution.Clone()
              });
            }
          }
          if (seedingStrategyLocal == SeedingStrategyTypes.NoSeeding) {
            lock (InstanceRuns) {
              InstanceRuns.Add(algorithmClone.Runs.Last());
            }
          } else {
            lock (SeededRuns) {
              SeededRuns.Add(algorithmClone.Runs.Last());
            }
          }
          waitHandle.Set();
        }
      };

      EventHandler<EventArgs<Exception>> exceptionOccurred = (sender, e) => {
        waitHandle.Set();
      };

      EventHandler timeChanged = (sender, e) => {
        IResult evalSolResult;
        if (!algorithmClone.Results.TryGetValue("EvaluatedSolutions", out evalSolResult) || !(evalSolResult.Value is Data.IntValue)) return;
        var evalSols = ((Data.IntValue)evalSolResult.Value).Value;
        if (evalSols >= MaximumEvaluations.Value && algorithmClone.ExecutionState == ExecutionState.Started)
          algorithmClone.Stop();
      };
      #endregion

      algorithmClone.ExecutionStateChanged += exeStateChanged;
      algorithmClone.ExceptionOccurred += exceptionOccurred;
      if (monitorStop) algorithmClone.ExecutionTimeChanged += timeChanged;

      return Task.Factory.StartNew(() => {
        algorithmClone.Start();
        OnAlgorithmInstanceStarted(algorithmClone);
        var cancelRequested = false;
        while (!waitHandle.WaitOne(200)) {
          if (cancellation.IsCancellationRequested) {
            cancelRequested = true;
            break;
          }
        }
        if (cancelRequested) {
          try { algorithmClone.Stop(); } catch { } // ignore race condition if it is stopped in the meantime
          waitHandle.WaitOne();
        }
        waitHandle.Dispose();
        return algorithmClone.Results;
      }, TaskCreationOptions.LongRunning);
    }

    public ResultCollection StartAlgorithm(int index, CancellationToken cancellation) {
      var task = StartAlgorithmAsync(index, cancellation);
      task.Wait(cancellation);
      return task.Result;
    }

    public Task UpdateKnowledgeBaseAsync(IProgress progress = null) {
      if (progress == null) progress = new Progress(string.Empty);
      progress.Start("Updating Knowledge Base from OKB");
      OnDownloadStarted(progress);
      return Task.Factory.StartNew(() => { DoUpdateKnowledgeBase(progress); }, TaskCreationOptions.LongRunning);
    }

    public void UpdateKnowledgeBase(IProgress progress = null) {
      UpdateKnowledgeBaseAsync(progress).Wait();
    }

    private void DoUpdateKnowledgeBase(IProgress progress) {
      var queryClient = Clients.OKB.Query.QueryClient.Instance;
      var adminClient = Clients.OKB.Administration.AdministrationClient.Instance;
      try {
        progress.Status = "Downloading run information...";
        progress.ProgressValue = 0;
        // FIXME: How to tell if refresh is necessary?
        queryClient.Refresh();
        progress.ProgressValue = 0.5;
        progress.Status = "Downloading algorithm and problem instance information...";
        // FIXME: How to tell if refresh is necessary?
        adminClient.Refresh();

        var probInstance = adminClient.Problems.SingleOrDefault(x => x.Id == Problem.ProblemId);
        if (probInstance == null) throw new InvalidOperationException("The chosen problem instance cannot be found in the OKB.");
        var probClassId = probInstance.ProblemClassId;

        var problemClassFilter = (Clients.OKB.Query.StringComparisonAvailableValuesFilter)queryClient.Filters.Single(x => x.Label == "Problem Class Name");
        problemClassFilter.Value = adminClient.ProblemClasses.Single(x => x.Id == probClassId).Name;

        progress.Status = "Downloading problem instances...";
        progress.ProgressValue = 0;
        var p = 0;
        ProblemInstances.UpdateOfRunsInProgress = true;
        ProblemInstances.Clear();
        var totalProblems = adminClient.Problems.Count(x => x.ProblemClassId == probClassId);
        foreach (var problInst in adminClient.Problems.Where(x => x.ProblemClassId == probClassId)) {
          progress.Status = string.Format("Downloading problem {0} (okb-id: {1})....", problInst.Name, problInst.Id);
          var data = Clients.OKB.Administration.AdministrationClient.GetProblemData(problInst.Id);
          if (data != null) {
            using (var stream = new MemoryStream(data)) {
              try {
                var prob = (IProblem)XmlParser.Deserialize<IContent>(stream);
                var probRun = new Run() { Name = prob.Name };
                prob.CollectParameterValues(probRun.Parameters);
                probRun.Parameters["Problem Name"] = new StringValue(prob.Name);
                probRun.Parameters["Problem Type"] = new StringValue(prob.GetType().Name);
                progress.Status += Environment.NewLine + "Downloading characteristics...";
                foreach (var v in RunCreationClient.Instance.GetCharacteristicValues(problInst.Id)) {
                  probRun.Results.Add("Characteristic." + v.Name, RunCreationClient.Instance.ConvertToItem(v));
                }
                ProblemInstances.Add(probRun);
              } catch { }
              stream.Close();
            }
          }
          p++;
          progress.ProgressValue = p / (double)totalProblems;
        }

        algorithmId2AlgorithmInstanceMapping.Clear();
        progress.Status = "Downloading algorithm instances...";
        progress.ProgressValue = 0;
        p = 0;
        foreach (var algInst in adminClient.Algorithms) {
          progress.Status = string.Format("Downloading algorithm {0} (okb-id: {1})...", algInst.Name, algInst.Id);
          var data = Clients.OKB.Administration.AdministrationClient.GetAlgorithmData(algInst.Id);
          if (data != null) {
            using (var stream = new MemoryStream(data)) {
              try {
                var alg = (IAlgorithm)XmlParser.Deserialize<IContent>(stream);
                algorithmId2AlgorithmInstanceMapping.Add(algInst.Id, alg);
              } catch { }
              stream.Close();
            }
          }
          p++;
          progress.ProgressValue = p / (double)adminClient.Algorithms.Count;
        }

        var interestingValues = queryClient.ValueNames.Where(x => InterestingValueNames.Contains(x.Name)).ToList();

        progress.Status = "Obtaining number of runs...";
        progress.ProgressValue = 0;
        p = 0;
        var count = queryClient.GetNumberOfRuns(problemClassFilter);
        if (count == 0) return;
        
        var runIds = queryClient.GetRunIds(problemClassFilter).ToList();
        var conversions = new List<Task>();
        var runList = new List<IRun>();
        while (p < count) {
          var nextIds = runIds.Skip(p).Take(500).ToList();
          progress.Status = string.Format("Downloading runs {0} to {1} of {2}...", p, p + nextIds.Count, count);
          var okbRuns = queryClient.GetRunsWithValues(nextIds, true, interestingValues);
          conversions.Add(Task.Factory.StartNew(() => {
            var hlRuns = okbRuns.AsParallel().Select(x => new { AlgorithmId = x.Algorithm.Id, Run = queryClient.ConvertToOptimizationRun(x) }).ToList();
            lock (runList) {
              foreach (var r in hlRuns) {
                algorithmId2RunMapping.Add(r.AlgorithmId, r.Run);
                runList.Add(r.Run);
              }
            }
          }));
          p += nextIds.Count;
          progress.ProgressValue = p / (double)count;
        }
        Task.WaitAll(conversions.ToArray());

        progress.Status = "Finishing...";
        var algInstRunDict = runList.Where(x => x.Parameters.ContainsKey("Problem Name") && x.Parameters["Problem Name"] is StringValue)
                                          .GroupBy(x => ((StringValue)x.Parameters["Problem Name"]).Value)
                                          .ToDictionary(x => x.Key, x => x.GroupBy(y => ((StringValue)y.Parameters["Algorithm Name"]).Value)
                                                                                  .ToDictionary(y => y.Key, y => y.ToList()));

        foreach (var instance in ProblemInstances) {
          IItem probNameParam;
          if (!instance.Parameters.TryGetValue("Problem Name", out probNameParam)) continue;

          var probInstanceName = ((StringValue)probNameParam).Value;
          var maximization = ((BoolValue)instance.Parameters["Maximization"]).Value;
          
          IItem bkParam;
          if (!instance.Parameters.TryGetValue("BestKnownQuality", out bkParam) || !(bkParam is DoubleValue)) {
            Dictionary<string, List<IRun>> algRuns;
            if (!algInstRunDict.TryGetValue(probInstanceName, out algRuns)) continue;
            var list = algRuns.SelectMany(x => x.Value)
                              .Where(x => x.Results.ContainsKey("QualityPerEvaluations"))
                              .Select(x => ((IndexedDataTable<double>)x.Results["QualityPerEvaluations"]).Rows.First().Values.Last().Item2);
            bkParam = new DoubleValue(maximization ? list.Max() : list.Min());
            instance.Parameters["BestKnownQuality"] = bkParam;
          } else bkParam = instance.Parameters["BestKnownQuality"];

          var bkQuality = ((DoubleValue)bkParam).Value;

          if (!algInstRunDict.ContainsKey(probInstanceName)) continue;
          foreach (var kvp in algInstRunDict[probInstanceName]) {
            // TODO: Things needs to be configurable here (table name, targets)
            foreach (var target in new[] { 1, 1.01, 1.05, 1.1, 1.2, 1.5 }) {
              var result = ExpectedRuntimeHelper.CalculateErt(kvp.Value, "QualityPerEvaluations", bkQuality * target, maximization);
              var resultName = kvp.Key + "@" + ((target - 1) * 100) + "%";
              IItem item;
              if (instance.Results.TryGetValue(resultName, out item)) {
                ((DoubleValue)item).Value = Math.Log10(result.ExpectedRuntime);
              } else instance.Results.Add(resultName, new DoubleValue(Math.Log10(result.ExpectedRuntime)));
            }
          }
        }
        try {
          KnowledgeBase.UpdateOfRunsInProgress = true;
          KnowledgeBase.Clear();
          KnowledgeBase.AddRange(runList);
        } finally { KnowledgeBase.UpdateOfRunsInProgress = false; }
      } finally { progress.Finish(); ProblemInstances.UpdateOfRunsInProgress = false; }
      UpdateInstanceProjection();
    }

    private void UpdateSuggestions() {
      if (Problem == null) return;
      var instances = new SortedList<double, IAlgorithm>();
      foreach (var relevantRuns in knowledgeBase.GroupBy(x => algorithmId2RunMapping.GetBySecond(x).Single())) {
        var algorithm = algorithmId2AlgorithmInstanceMapping.GetByFirst(relevantRuns.Key);
        var avgQuality = 0.0;
        var counter = 0;
        foreach (var problemRuns in relevantRuns.GroupBy(x => ((StringValue)x.Parameters["Problem Name"]).Value)) {
          var probInstance = ProblemInstances.SingleOrDefault(x => x.Name == problemRuns.Key);
          if (probInstance == null) continue;
          var bkQuality = ((DoubleValue)probInstance.Parameters["BestKnownQuality"]).Value;
          foreach (var run in problemRuns) {
            var performanceGraph = ((IndexedDataTable<double>)run.Results["QualityPerEvaluations"]);
            try {
              avgQuality += performanceGraph.Rows.First().Values.TakeWhile(x => x.Item1 < MaximumEvaluations.Value).Last().Item2 / bkQuality;
              counter++;
            } catch {
              continue;
            }
          }
        }
        avgQuality /= counter;
        instances.Add(avgQuality, algorithm);
      }

      var instanceLadder = instances.Select(x => (IAlgorithm)x.Value.Clone()).ToList();
      if (Maximization) instanceLadder.Reverse();
      suggestedInstances.Replace(instanceLadder);
    }

    public event EventHandler<EventArgs<IProgress>> DownloadStarted;
    private void OnDownloadStarted(IProgress progress) {
      var handler = DownloadStarted;
      if (handler != null) handler(this, new EventArgs<IProgress>(progress));
    }

    public event EventHandler<EventArgs<IAlgorithm>> AlgorithmInstanceStarted;
    private void OnAlgorithmInstanceStarted(IAlgorithm instance) {
      var handler = AlgorithmInstanceStarted;
      if (handler != null) handler(this, new EventArgs<IAlgorithm>(instance));
    }
  }
}
